@@ -12,25 +12,44 @@ import time
 import urllib
 
 import dateparser
+import dateparser.search
 import puz
 import requests
+import yaml
+
+from getpass import getpass
 
 from bs4 import BeautifulSoup
 from html2text import html2text
+
+# This imports the _module_ unidecode, which converts Unicode strings to
+# plain ASCII. The puz format, however, can accept Latin1, which is a larger
+# subset. So the second line tells the module to leave codepoints 128-256
+# untouched, then we import the _function_ unidecode.
+import unidecode
+unidecode.Cache[0] = [chr(c) if c > 127 else '' for c in range(256)]
 from unidecode import unidecode
 
-__version__ = '2021.1.3'
+__version__ = '2022.05.21'
+CONFIG_PATH = os.environ.get('XDG_CONFIG_HOME') or os.path.expanduser('~/.config')
+CONFIG_PATH = os.path.join(CONFIG_PATH, 'xword-dl/xword-dl.yaml')
+
+if not os.path.exists(CONFIG_PATH):
+    os.makedirs(os.path.dirname(CONFIG_PATH), exist_ok=True)
+    open(CONFIG_PATH, 'a').close()
 
 
-def by_keyword(keyword, date=None, filename=None):
+def by_keyword(keyword, **kwargs):
     keyword_dict = {d[1].command: d[1] for d in get_supported_outlets()}
     selected_downloader = keyword_dict.get(keyword, None)
 
     if selected_downloader:
-        dl = selected_downloader()
+        dl = selected_downloader(**kwargs)
     else:
         raise ValueError('Keyword {} not recognized.'.format(keyword))
 
+    date = kwargs.get('date')
+
     if not date:
         puzzle_url = dl.find_latest()
     elif date and hasattr(dl, 'find_by_date'):
@@ -42,8 +61,7 @@ def by_keyword(keyword, date=None, filename=None):
             'Selection by date not available for {}.'.format(dl.outlet))
 
     puzzle = dl.download(puzzle_url)
-
-    filename = filename or dl.pick_filename(puzzle)
+    filename = dl.pick_filename(puzzle, filename=kwargs.get('filename'))
 
     return puzzle, filename
 
@@ -61,7 +79,7 @@ def by_url(url, filename=None):
                                  if site[0] in netloc), None)
 
     if supported_downloader:
-        dl = supported_downloader()
+        dl = supported_downloader(netloc=netloc)
         puzzle_url = url
     else:
         amuse_url = None
@@ -75,7 +93,7 @@ def by_url(url, filename=None):
                 break
 
         if amuse_url:
-            dl = AmuseLabsDownloader()
+            dl = AmuseLabsDownloader(netloc=netloc)
             puzzle_url = amuse_url
 
     if dl:
@@ -126,7 +144,7 @@ def remove_invalid_chars_from_filename(filename):
 
 
 def parse_date(entered_date):
-    return dateparser.parse(entered_date)
+    return dateparser.parse(entered_date, settings={'PREFER_DATES_FROM':'past'})
 
 
 def parse_date_or_exit(entered_date):
@@ -139,25 +157,80 @@ def parse_date_or_exit(entered_date):
     return guessed_dt
 
 
+def update_config_file(heading, new_values_dict):
+    with open(CONFIG_PATH, 'r') as f:
+        config = yaml.safe_load(f) or {}
+
+    if heading not in config:
+        config[heading] = {}
+
+    config[heading].update(new_values_dict)
+
+    with open(CONFIG_PATH, 'w') as f:
+        yaml.dump(config, f)
+
+
 class BaseDownloader:
     outlet = None
     outlet_prefix = None
 
-    def __init__(self):
+    def __init__(self, **kwargs):
         self.date = None
+        self.netloc = None
+
+        self.settings = {}
+
+        with open(CONFIG_PATH, 'r') as f:
+            config = yaml.safe_load(f) or {}
+
+        self.settings.update(config.get('general', {}))
+        if hasattr(self, 'command'):
+            self.settings.update(config.get(self.command, {}))
+        elif 'netloc' in kwargs:
+            self.netloc = kwargs['netloc']
+            self.settings.update(config.get('url', {}))
+            self.settings.update(config.get(self.netloc, {}))
+        self.settings.update(kwargs)
 
     def pick_filename(self, puzzle, **kwargs):
+        tokens = {'outlet':  self.outlet or '',
+                  'prefix':  self.outlet_prefix or '',
+                  'title':   puzzle.title or '',
+                  'author':  puzzle.author or '',
+                  'cmd':     (self.command if hasattr(self, 'command')
+                              else self.netloc or ''),
+                  'netloc':  self.netloc or '',
+                 }
+
+        tokens = {t:kwargs[t] if t in kwargs else tokens[t] for t in tokens}
+
+        date = kwargs.get('date', self.date)
+
+        template = kwargs.get('filename') or self.settings.get('filename') or ''
+
+        if not template:
+            template += '%prefix' if tokens.get('prefix') else '%author'
+            template += ' - %Y%m%d' if date  else ''
+            template += ' - %title' if tokens.get('title') else ''
+
+        for token in tokens.keys():
+            replacement = (kwargs.get(token) if token in kwargs
+                           else tokens[token])
+            template = template.replace('%' + token, replacement)
+
+
+        if date:
+            template = date.strftime(template)
+
         title = kwargs.get('title', puzzle.title)
         date = kwargs.get('date', self.date)
 
-        date = date.strftime('%Y%m%d') if date else None
+        if not template.endswith('.puz'):
+            template += '.puz'
 
-        filename_components = [component for component in
-                               [self.outlet_prefix or puzzle.author,
-                                date,
-                                title] if component]
+        filename = remove_invalid_chars_from_filename(template)
 
-        return " - ".join(filename_components) + '.puz'
+        return filename
 
     def find_solver(self, url):
         """Given a URL for a puzzle, returns the essential 'solver' URL.
@@ -196,7 +269,7 @@ class BaseDownloader:
 
 class AmuseLabsDownloader(BaseDownloader):
     def __init__(self, **kwargs):
-        super().__init__()
+        super().__init__(**kwargs)
 
         self.id = None
 
@@ -209,8 +282,25 @@ class AmuseLabsDownloader(BaseDownloader):
             puzzle_ids = puzzles.findAll('li', attrs={'class': 'tile'})
         self.id = puzzle_ids[0].get('data-id', '')
 
+        self.get_and_add_picker_token(res.text)
+
         return self.find_puzzle_url_from_id(self.id)
 
+    def get_and_add_picker_token(self, picker_source=None):
+        if not picker_source:
+            res = requests.get(self.picker_url)
+            picker_source = res.text
+
+        rawsps = next((line.strip() for line in picker_source.splitlines()
+                     if 'pickerParams.rawsps' in line), None)
+
+        if rawsps:
+            rawsps = rawsps.split("'")[1]
+            picker_params = json.loads(base64.b64decode(rawsps).decode("utf-8"))
+            token = picker_params.get('pickerToken', None)
+            if token:
+                self.url_from_id += '&pickerToken=' + token
+
     def find_puzzle_url_from_id(self, puzzle_id):
         return self.url_from_id.format(puzzle_id=puzzle_id)
 
@@ -237,15 +327,37 @@ class AmuseLabsDownloader(BaseDownloader):
 
         rawc = rawc.split("'")[1]
 
-        xword_data = json.loads(base64.b64decode(rawc).decode("utf-8"))
+        # helper function to decode rawc
+        # as occasionally it can be obfuscated
+        def load_rawc(rawc):
+            if '.' not in rawc:
+                return json.loads(base64.b64decode(rawc).decode("utf-8"))
+            rawcParts = rawc.split(".")
+            buff = list(rawcParts[0])
+            key1 = rawcParts[1][::-1]
+            key = [int(k, 16) + 2 for k in key1]
+            i, segmentCount = (0, 0)
+            while i < len(buff) - 1:
+                # reverse sections of the buffer, using key digits as lengths
+                segmentLength = min(key[segmentCount % len(key)], len(buff) - i)
+                for j in range(segmentLength // 2):
+                    buff[i+j], buff[i + segmentLength - j - 1] = (
+                               buff[i + segmentLength - j - 1], buff[i+j])
+                i += segmentLength
+                segmentCount += 1
+
+            newRawc = ''.join(buff)
+            return json.loads(base64.b64decode(newRawc).decode("utf-8"))
+
+        xword_data = load_rawc(rawc)
 
         return xword_data
 
     def parse_xword(self, xword_data):
         puzzle = puz.Puzzle()
-        puzzle.title = xword_data.get('title', '').strip()
-        puzzle.author = xword_data.get('author', '').strip()
-        puzzle.copyright = xword_data.get('copyright', '').strip()
+        puzzle.title = unidecode(xword_data.get('title', '').strip())
+        puzzle.author = unidecode(xword_data.get('author', '').strip())
+        puzzle.copyright = unidecode(xword_data.get('copyright', '').strip())
         puzzle.width = xword_data.get('w')
         puzzle.height = xword_data.get('h')
 
@@ -296,7 +408,7 @@ class AmuseLabsDownloader(BaseDownloader):
 
         clues = [word['clue']['clue'] for word in weirdass_puz_clue_sorting]
 
-        normalized_clues = [html2text(unidecode(clue), bodywidth=0)
+        normalized_clues = [html2text(unidecode(clue), bodywidth=0).strip()
                             for clue in clues]
         puzzle.clues.extend(normalized_clues)
 
@@ -361,6 +473,9 @@ class AtlanticDownloader(AmuseLabsDownloader):
 
         return self.find_puzzle_url_from_id(self.id)
 
+    def guess_date_from_id(self, puzzle_id):
+        self.date = datetime.datetime.strptime(puzzle_id.split('_')[1],
+                                               '%Y%m%d')
 
 class NewsdayDownloader(AmuseLabsDownloader):
     command = 'nd'
@@ -403,16 +518,18 @@ class LATimesDownloader(AmuseLabsDownloader):
         url_formatted_date = dt.strftime('%y%m%d')
         self.id = 'tca' + url_formatted_date
 
+        self.get_and_add_picker_token()
+
         return self.find_puzzle_url_from_id(self.id)
 
     def pick_filename(self, puzzle, **kwargs):
         split_on_dashes = puzzle.title.split(' - ')
         if len(split_on_dashes) > 1:
-            use_title = split_on_dashes[-1].strip()
+            title = split_on_dashes[-1].strip()
         else:
-            use_title = ''
+            title = ''
 
-        return super().pick_filename(puzzle, title=use_title)
+        return super().pick_filename(puzzle, title=title, **kwargs)
 
 
 class NewYorkerDownloader(AmuseLabsDownloader):
@@ -487,7 +604,50 @@ class NewYorkerDownloader(AmuseLabsDownloader):
                 title = main.strip()
         except ValueError:
             title = puzzle.title
-        return super().pick_filename(puzzle, title=title)
+        return super().pick_filename(puzzle, title=title, **kwargs)
+
+
+class VoxDownloader(AmuseLabsDownloader):
+    command = 'vox'
+    outlet = 'Vox'
+    outlet_prefix = 'Vox'
+
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+
+        self.picker_url = 'https://cdn3.amuselabs.com/vox/date-picker?set=vox'
+        self.url_from_id = 'https://cdn3.amuselabs.com/vox/crossword?id={puzzle_id}&set=vox'
+
+    def guess_date_from_id(self, puzzle_id):
+        self.date = datetime.datetime.strptime(puzzle_id.split('_')[1],
+                                               '%Y%m%d')
+
+class DailyBeastDownloader(AmuseLabsDownloader):
+    command = 'db'
+    outlet = 'Daily Beast'
+    outlet_prefix = 'Daily Beast'
+
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+
+        self.picker_url = 'https://cdn3.amuselabs.com/tdb/date-picker?set=tdb'
+        self.url_from_id = 'https://cdn3.amuselabs.com/tdb/crossword?id={puzzle_id}&set=tdb'
+
+    def parse_xword(self, xword_data):
+        puzzle = super().parse_xword(xword_data)
+
+        # Daily Beast puzzle IDs, unusually for AmuseLabs puzzles, don't include
+        # the date. This pulls it out of the puzzle title, which will work
+        # as long as that stays consistent.
+
+        possible_dates = dateparser.search.search_dates(puzzle.title)
+
+        if possible_dates:
+            self.date = possible_dates[-1][1]
+        else:
+            self.date = datetime.datetime.today()
+
+        return puzzle
 
 
 class WSJDownloader(BaseDownloader):
@@ -528,7 +688,7 @@ class WSJDownloader(BaseDownloader):
             return self.find_solver(puzzle_link)
 
     def fetch_data(self, solver_url):
-        data_url = solver_url.replace('index.html', 'data.json')
+        data_url = solver_url.replace('?embed=1', 'data.json')
         return requests.get(data_url, headers=self.headers).json()['data']
 
     def parse_xword(self, xword_data):
@@ -540,7 +700,7 @@ class WSJDownloader(BaseDownloader):
         self.date = datetime.datetime.strptime(date_string, '%Y/%m/%d')
 
         fetched = {}
-        for field in ['title', 'byline', 'publisher']:
+        for field in ['title', 'byline', 'publisher', 'description']:
             fetched[field] = html2text(xword_metadata.get(field, ''),
                                        bodywidth=0).strip()
 
@@ -551,6 +711,8 @@ class WSJDownloader(BaseDownloader):
         puzzle.width = int(xword_metadata.get('gridsize').get('cols'))
         puzzle.height = int(xword_metadata.get('gridsize').get('rows'))
 
+        puzzle.notes = fetched.get('description')
+
         solution = ''
         fill = ''
         markup = b''
@@ -578,7 +740,7 @@ class WSJDownloader(BaseDownloader):
 
         clues = [clue['clue'] for clue in sorted_clue_list]
         normalized_clues = [
-            html2text(unidecode(clue), bodywidth=0) for clue in clues]
+            html2text(unidecode(clue), bodywidth=0).strip() for clue in clues]
 
         puzzle.clues = normalized_clues
 
@@ -594,7 +756,7 @@ class WSJDownloader(BaseDownloader):
 
 class AMUniversalDownloader(BaseDownloader):
     def __init__(self, **kwargs):
-        super().__init__()
+        super().__init__(**kwargs)
         self.url_blob = None
 
     def find_by_date(self, dt):
@@ -614,15 +776,7 @@ class AMUniversalDownloader(BaseDownloader):
         attempts = 3
         while attempts:
             try:
-                fullpath = os.path.abspath(os.path.dirname(__file__))
-                certpath = os.path.join(fullpath,
-                                        'cert/embed-universaluclick-com-chain.pem')
-                res = requests.get(solver_url, verify=certpath)
-                # That cert is required because UUclick has a misconfigured
-                # certificate chain. Ideally they will fix that and this part
-                # can be removed.
-                # NOTE: the cert issue doesn't affect USA Today, but including
-                # the cert here doesn't break anything.
+                res = requests.get(solver_url)
                 xword_data = res.json()
                 break
             except json.JSONDecodeError:
@@ -689,7 +843,7 @@ class USATodayDownloader(AMUniversalDownloader):
     outlet_prefix = 'USA Today'
 
     def __init__(self, **kwargs):
-        super().__init__()
+        super().__init__(**kwargs)
 
         self.url_blob = 'https://gamedata.services.amuniversal.com/c/uupuz/l/U2FsdGVkX18CR3EauHsCV8JgqcLh1ptpjBeQ%2Bnjkzhu8zNO00WYK6b%2BaiZHnKcAD%0A9vwtmWJp2uHE9XU1bRw2gA%3D%3D/g/usaon/d/'
 
@@ -705,11 +859,166 @@ class UniversalDownloader(AMUniversalDownloader):
     outlet_prefix = 'Universal'
 
     def __init__(self, **kwargs):
-        super().__init__()
+        super().__init__(**kwargs)
 
         self.url_blob = 'https://embed.universaluclick.com/c/uucom/l/U2FsdGVkX18YuMv20%2B8cekf85%2Friz1H%2FzlWW4bn0cizt8yclLsp7UYv34S77X0aX%0Axa513fPTc5RoN2wa0h4ED9QWuBURjkqWgHEZey0WFL8%3D/g/fcx/d/'
 
 
+class NewYorkTimesDownloader(BaseDownloader):
+    command = 'nyt'
+    outlet = 'New York Times'
+    outlet_prefix = 'NY Times'
+
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+
+        self.url_from_id = 'https://www.nytimes.com/svc/crosswords/v2/puzzle/{}.json'
+
+        self.headers = {}
+        self.cookies = {}
+
+        username = self.settings.get('username')
+        password = self.settings.get('password')
+
+        if username and password:
+            nyts_token = self.authenticate(username, password)
+            update_config_file('nyt', {'NYT-S': nyts_token})
+        else:
+            nyts_token = self.settings.get('NYT-S')
+
+        if not nyts_token:
+            raise ValueError('No credentials provided or stored. Try running xword-dl nyt --authenticate')
+        else:
+            self.cookies.update({'NYT-S': nyts_token})
+
+    def authenticate(self, username, password):
+        """Given a NYT username and password, returns the NYT-S cookie value"""
+
+        res = requests.post('https://myaccount.nytimes.com/svc/ios/v2/login',
+                data={'login': username, 'password': password},
+                headers={'User-Agent':
+                    'Crossword/20211014193428 CFNetwork/1240.0.4 Darwin/20.6.0',
+                    'client_id': 'ios.crosswords',})
+
+        res.raise_for_status()
+
+        nyts_token = ''
+
+        for cookie in res.json()['data']['cookies']:
+            if cookie['name'] == 'NYT-S':
+                nyts_token = cookie['cipheredValue']
+
+        if nyts_token:
+            return nyts_token
+        else:
+            raise ValueError('NYT-S cookie not found.')
+
+    def find_latest(self):
+        oracle = "https://www.nytimes.com/svc/crosswords/v2/oracle/daily.json"
+
+        res = requests.get(oracle)
+        puzzle_id = res.json()['results']['current']['puzzle_id']
+
+        url = self.url_from_id.format(puzzle_id)
+
+        return url
+
+    def find_by_date(self, dt):
+        lookup_url = 'https://www.nytimes.com/svc/crosswords/v3/puzzles.json?status=published&order=published&sort=asc&pad=false&print_date_start={}&print_date_end={}&publish_type=daily'
+
+        formatted_date = dt.strftime('%Y-%m-%d')
+
+        res = requests.get(lookup_url.format(formatted_date, formatted_date))
+
+        puzzle_id = res.json()['results'][0]['puzzle_id']
+
+        return self.url_from_id.format(puzzle_id)
+
+    def find_solver(self, url):
+        return url
+
+    def fetch_data(self, solver_url):
+        res = requests.get(solver_url, cookies=self.cookies)
+        res.raise_for_status()
+
+        return res.json()['results'][0]
+
+    def parse_xword(self, xword_data):
+        puzzle = puz.Puzzle()
+
+        metadata = xword_data.get('puzzle_meta')
+        puzzle.author = metadata.get('author').strip()
+        puzzle.copyright = metadata.get('copyright').strip()
+        puzzle.height = metadata.get('height')
+        puzzle.width = metadata.get('width')
+
+        if metadata.get('notes'):
+            puzzle.notes = metadata.get('notes')[0]['txt'].strip()
+
+        date_string = metadata.get('printDate')
+
+        self.date = datetime.datetime.strptime(date_string, '%Y-%m-%d')
+
+        puzzle.title = metadata.get('title') or self.date.strftime(
+                '%A, %B %d, %Y')
+
+        puzzle_data = xword_data['puzzle_data']
+
+        solution = ''
+        fill = ''
+        markup = b''
+        rebus_board = []
+        rebus_index = 0
+        rebus_table = ''
+
+        for idx, square in enumerate(puzzle_data['answers']):
+            if not square:
+                solution += '.'
+                fill += '.'
+                rebus_board.append(0)
+            elif len(square) == 1:
+                solution += square
+                fill += '-'
+                rebus_board.append(0)
+            else:
+                solution += square[0][0]
+                fill += '-'
+                rebus_board.append(rebus_index + 1)
+                rebus_table += '{:2d}:{};'.format(rebus_index, square[0])
+                rebus_index += 1
+
+            markup += (b'\x80' if puzzle_data['layout'][idx] == 3 else b'\x00')
+
+        puzzle.solution = solution
+        puzzle.fill = fill
+
+        clue_list = puzzle_data['clues']['A'] + puzzle_data['clues']['D']
+        clue_list.sort(key=lambda c: c['clueNum'])
+
+        puzzle.clues = [unidecode(c['value']).strip() for c in clue_list]
+
+        if b'\x80' in markup:
+            puzzle.extensions[b'GEXT'] = markup
+            puzzle._extensions_order.append(b'GEXT')
+            puzzle.markup()
+
+        if any(rebus_board):
+            puzzle.extensions[b'GRBS'] = bytes(rebus_board)
+            puzzle.extensions[b'RTBL'] = rebus_table.encode(puz.ENCODING)
+            puzzle._extensions_order.extend([b'GRBS', b'RTBL'])
+            puzzle.rebus()
+
+        return puzzle
+
+    def pick_filename(self, puzzle, **kwargs):
+        if puzzle.title == self.date.strftime('%A, %B %d, %Y'):
+            title = ''
+        else:
+            title = puzzle.title
+
+        return super().pick_filename(puzzle, title=title, **kwargs)
+
+
 def main():
     parser = argparse.ArgumentParser(prog='xword-dl',
                                      description=textwrap.dedent("""\
@@ -745,32 +1054,74 @@ def main():
                           help='a specific puzzle date to select',
                           default=[])
 
+    parser.add_argument('-a', '--authenticate',
+                        help=textwrap.dedent("""\
+                            when used with the nyt puzzle keyword,
+                            stores an authenticated New York Times cookie
+                            without downloading a puzzle. If username
+                            or password are not provided as flags,
+                            xword-dl will prompt for those values
+                            at runtime"""),
+                        action='store_true',
+                        default=False)
+
+    parser.add_argument('-u', '--username',
+                        help=textwrap.dedent("""\
+                            username for a site that requires credentials
+                            (currently only the New York Times)"""),
+                        default=None)
+
+    parser.add_argument('-p', '--password',
+                        help=textwrap.dedent("""\
+                            password for a site that requires credentials
+                            (currently only the New York Times)"""),
+                        default=None)
+
     parser.add_argument('-o', '--output',
                         help=textwrap.dedent("""\
                             the filename for the saved puzzle
                             (if not provided, a default value will be used)"""),
                         default=None)
 
+
     args = parser.parse_args()
+    if args.authenticate and args.source == 'nyt':
+        username = args.username or input("New York Times username: ")
+        password = args.password or getpass("Password: ")
+
+        try:
+            dl = NewYorkTimesDownloader(username=username, password=password)
+            sys.exit('Authentication successful.')
+        except Exception as e:
+            sys.exit(' '.join(['Authentication failed:', str(e)]))
+    elif args.authenticate:
+        sys.exit('Authentication flag must use a puzzle outlet keyword.')
+
     if not args.source:
         sys.exit(parser.print_help())
 
+    options = {}
+    if args.username:
+        options['username'] = args.username
+    if args.password:
+        options['password'] = args.password
+    if args.output:
+        options['filename'] = args.output
+    if args.date:
+        options['date'] = ''.join(args.date)
+
     try:
         if args.source.startswith('http'):
             puzzle, filename = by_url(args.source,
                                       filename=args.output)
         else:
-            puzzle, filename = by_keyword(args.source,
-                                          date=''.join(args.date),
-                                          filename=args.output)
+            puzzle, filename = by_keyword(args.source, **options)
     except ValueError as e:
         sys.exit(e)
 
     if not filename.endswith('.puz'):
         filename = filename + '.puz'
 
-    filename = remove_invalid_chars_from_filename(filename)
-
     save_puzzle(puzzle, filename)
 
 
